<?php

// TPM_Keepass_Import.php v.1.6
// http://teampasswordmanager.com/docs/keepass-import/
//
// Ferran Barba, June 2017
// info@teampasswordmanager.com
// http://teampasswordmanager.com
//
// PHP class to import projects and passwords from Keepass XML 2.x Files into Team Password Manager

class TPM_Keepass_Import {

	// API version
	const API_VERSION = 4;

	// Keepass XML 2.x file to import
	protected $_file = '';

	// URL of the installation of Team Password Manager, including index.php
	// Example: http://mytpm.mydomain.com/index.php/
	protected $_tpm_base_url = '';

	// TPM credentials
	protected $_tpm_username = '';
	protected $_tpm_password = '';

	// Test Mode: will only output data to the log, will not import anything to TPM
	// In this mode, the TPM url and credentials are not needed
	protected $_test_mode = FALSE;

	// Parent project (internal ID in TPM): imported projects will be childs of this project (0=root)
	protected $_parent_project = 0;

	// Log handle
	protected $_log_handle = NULL; 

	// Handle for XML file
	protected $_reader = NULL;

	// **************** CONSTRUCTOR AND PARAMETERS ****************

	public function __construct ( $file = '', $tpm_base_url = '', $tpm_username = '', $tpm_password = '', $parent_project=0 ) {
		if ( $file ) {
			$this->set_file($file);
		}

		if ( $tpm_base_url && $tpm_username && $tpm_password ) {
			$this->set_tpm_credentials($tpm_base_url, $tpm_username, $tpm_password);
		}

		$this->set_parent_project($parent_project);
	}

	public function set_file ( $file ) {
		$this->_file = $file;
	}

	public function set_tpm_credentials ( $tpm_base_url, $tpm_username, $tpm_password ) {
		$this->_tpm_base_url = $tpm_base_url;
		if ( substr($this->_tpm_base_url, strlen($this->_tpm_base_url) - 1) != '/' ) {
			$this->_tpm_base_url .= '/';
		}

		$this->_tpm_username = $tpm_username;
		$this->_tpm_password = $tpm_password;
	}	

	public function set_parent_project ( $parent_project ) {
		$this->_parent_project = $parent_project;
	}

	// **************** LOG ****************

	protected function _open_log ( ) {
		$filepath = 'tpm_import_log_' . date("Ymd_His") . '.log';
		$this->_log_handle = fopen($filepath, 'a');
		if ( ! $this->_log_handle ) {
			$this->_tpm_die('Failed to create log file.');
		}
	}

	protected function _close_log ( ) {
		if ( $this->_log_handle ) {	
			fclose($this->_log_handle);
		}
	}

	protected function _write_log ( $str_to_write ) {
		$log_string = 'date_time=' . date("Y-m-d H:i:s") . ' ' . $str_to_write . "\n";
		if ( $this->_log_handle ) {	
			fwrite($this->_log_handle, $log_string);
		}
	}

	// **************** IMPORT ****************

	public function import ( $test = FALSE, $pretty_print_projects = FALSE ) {
		// Some vars
		$inside = ''; // Element currently reading: group, entry
		$element_to_add = array(); // Contains the data for the current element being read from the xml file (project/group or password/entry)
		$current_depth = 0;
		$initial_depth = 0;
		$current_project_id = $this->_parent_project; // Last added project (or initial)
		$current_parent_id = $this->_parent_project; 
		$projects_stack = array();
		$rename_counter = 1; // If we find identical names in projects/passwords we add a suffix: REN+$rename_counter

		// Open file and log
		$this->_open_xml();
		$this->_open_log();
		$this->_write_log('keepass_file=' . $this->_file);
		if ( $test ) {
			$this->_write_log('begin_test');
		} else {
			$this->_write_log('begin_import');
			$this->_write_log('tpm_base_url=' . $this->_tpm_base_url);
			$this->_write_log('tpm_username=' . $this->_tpm_username);
			$this->_write_log('parent_project_id=' . $this->_parent_project);
		}

		// Move to the first <Group /> node
		while ($this->_reader->read() && $this->_reader->name != 'Group');

		if ( $this->_reader->name == 'Group' ) {
			$inside = 'group';
			$current_depth = $this->_reader->depth;
			$previous_depth = 1; // start with this depth
			$element_to_add = array();
		}

		// Main loop
		while ($this->_reader->read()) { 
			$xml_element = $this->_reader->name; 

			if (  $xml_element == 'Group' ) { // We've found a Group (so a new project)
				if ( $inside == 'group' ) { // We were inside a Group (so project)
					$depth_to_add = $current_depth;
					$current_depth = $this->_reader->depth;

					// Log and import the previously read project (element_to_add)
					$this->_log_import_project($element_to_add, $projects_stack, $current_project_id, $current_parent_id, $previous_depth, $depth_to_add, $test, $pretty_print_projects, $rename_counter);
				} // inside=group
				else {
					$this->_log_import_password($element_to_add, $current_project_id, $depth_to_add, $test, $pretty_print_projects, $rename_counter);
				}
				
				$inside = 'group';

			} elseif ( $xml_element == 'Entry' ) {

				if ( $inside == 'group' ) { // We were inside a Group (so project)
					$depth_to_add = $current_depth;
					$current_depth = $this->_reader->depth;

					// Log and import the previously read project (element_to_add)
					$this->_log_import_project($element_to_add, $projects_stack, $current_project_id, $current_parent_id, $previous_depth, $depth_to_add, $test, $pretty_print_projects, $rename_counter);
				} // inside=group
				else {
					$this->_log_import_password($element_to_add, $current_project_id, $depth_to_add, $test, $pretty_print_projects, $rename_counter);
				}

				$inside = 'entry';

			} elseif ( $xml_element == 'History' ) {
				// Skip all history
				$this->_reader->next();
			} else {
				if ( $inside == 'group' ) {
					if ( $this->_reader->nodeType == XMLReader::ELEMENT ) {
		            	if ( $xml_element == 'UUID' || $xml_element == 'Name' || $xml_element == 'Notes' ) {
		            		$this->_reader->read();
		            		$element_to_add[$xml_element] = trim($this->_reader->value);
						}
		        	}		
				} elseif ( $inside == 'entry' ) {
					if ( $this->_reader->nodeType == XMLReader::ELEMENT ) {
		            	if ( $xml_element == 'UUID' || $xml_element == 'Tags' ) {
		            		$this->_reader->read();
		            		$rvalue = trim($this->_reader->value);
		            		if ( $xml_element == 'Tags' ) {
		            			// replace ; with ,
		            			$rvalue = str_replace(';', ',', $rvalue);
		            		}
		            		$element_to_add[$xml_element] = $rvalue;
						} elseif ( $xml_element == 'String' ) { // data, including custom fields
							// Key / Value
							$this->_reader->read();
							$this->_reader->read();
							if ( $this->_reader->name == 'Key' ) {
								$this->_reader->read();
								$pwd_field = $this->_reader->value;
								$this->_reader->read();								
								$this->_reader->read();
								$this->_reader->read();
								if ( $this->_reader->name == 'Value' ) {
									$this->_reader->read();
									$element_to_add[$pwd_field] = trim($this->_reader->value);
								}
							}
						} elseif ( $xml_element == 'Times' ) { // expiry date
							while ($this->_reader->read() && $this->_reader->name != 'ExpiryTime');
							$this->_reader->read();
							$edate = $this->_reader->value; // ExpiryDate content
							while ($this->_reader->read() && $this->_reader->name != 'Expires');
							$this->_reader->read();
							if ( $this->_reader->value == 'True' ) {
								$element_to_add['expiry_date'] = date('Y-m-d',strtotime($edate));
							}							
						} // which node (UUID, Tags, String, Times)
		        	} // nodeType == XMLReader::ELEMENT
				 } // inside group / entry
			} // $xml_element (Group, Entry, History, other)
		} // main while

		// Close
		if ( $test ) {
			$this->_write_log('end_test');
		} else {
			$this->_write_log('end_import');
		}
		$this->_close_log();
		$this->_close_xml();
	} // import()

	// Logs and/or imports the project
	protected function _log_import_project ( &$element_to_add, &$projects_stack, 
											 &$current_project_id, &$current_parent_id, 
											 &$previous_depth, $depth_to_add, 
											 $test, $pretty_print_projects, &$rename_counter ) {
		if ( $element_to_add ) {						
			$str_log_line = '';

			// If Name is not found, we add it
			if ( !isset($element_to_add['Name']) || !$element_to_add['Name'] ) {
				$element_to_add['Name'] = 'No name ' . '#REN' . $rename_counter . '#';
				$rename_counter++;
			}

			if ( $pretty_print_projects ) {
				$str_log_line = '@';
				for ( $i=1; $i<=($depth_to_add-2); $i++ ) {
					$str_log_line .= '--';
				}
				$str_log_line .= $element_to_add['Name'];
			} else {
				$str_log_line .= 'project=';
				$str_log_line .= $element_to_add['Name'];
				$str_log_line .= ' depth=';
				$str_log_line .= $depth_to_add-2;
			}

			// Add if not test
			if ( ! $test ) {
				// current_parent_id:
				// 	- if depth_to_add == previous_depth => current_parent_id = the same
				// 	- if depth_to_add > previous_depth => current_parent_id = current_project_id AND push current parent into stack
				// 	- if depth_to_add < previous_depth => current_parent_id = pop the difference

				// If it's a subproject, the parent is the current project and push into the stack
				if ( $depth_to_add > $previous_depth ) {
					if ( $previous_depth > 1 ) {
						array_push($projects_stack, $current_parent_id);	
					}
					$current_parent_id = $current_project_id;
				} elseif ( $depth_to_add < $previous_depth ) {
					$adiff = $previous_depth-$depth_to_add;
					for ( $ai=1; $ai<=$adiff; $ai++ ) {
						$current_parent_id = array_pop($projects_stack);
					}
				}

				$renamed = '';
				$imported_id = $this->_add_project($element_to_add, $current_parent_id, $rename_counter, $renamed);

				$current_project_id = $imported_id;

				if ( $renamed ) {
					$str_log_line .= ' ' . $renamed;					
				}
				$str_log_line .= ' imported_id=' . $imported_id;

				$previous_depth = $depth_to_add;		
			}

			$this->_write_log($str_log_line);

			// Delete data of the project
			$element_to_add = array();
		} // we have a project to add (element_to_add)
	} // _log_import_project()

	// Adds the project to TPM
	protected function _add_project ( $element_to_add, $parent_id, &$rename_counter, &$renamed ) {
		$rbprj = array(
			'name' => $element_to_add['Name'],
			'parent_id' => $parent_id,
			'notes'=> (isset($element_to_add['Notes']) ? $element_to_add['Notes'] : ''),
		);

		$rb = json_encode($rbprj);
		$ret = $this->_http_post('projects.json', $rb);

		$status = intval($ret[0]);
		$arr_rbody = json_decode($ret[1], TRUE);

		if ( $status == 201 ) {
			// Added, get the ID
			return intval($arr_rbody['id']);
		} else {
			if ( $status == 409 && $arr_rbody['message'] == 'This project already exists in this level' ) {
				// Let's add a suffix to the name and try it again
				$renamed = '#REN' . $rename_counter . '#';
				$rename_counter++;
				$rbprj['name'] = $element_to_add['Name'] . ' ' . $renamed;
				$rb = json_encode($rbprj);
				$ret = $this->_http_post('projects.json', $rb);

				$status = intval($ret[0]);
				$arr_rbody = json_decode($ret[1], TRUE);
				if ( $status == 201 ) {
					// Added, get the ID
					return intval($arr_rbody['id']);
				} else {
					$this->_tpm_die('Error adding project: ' . $status . '-' . $arr_rbody['message']);		
				}
			} else {	
				$this->_tpm_die('Error adding project: ' . $status . '-' . $arr_rbody['message']);	
			}
		}
	} // _add_project()

	// Logs and/or imports the password
	protected function _log_import_password ( &$element_to_add, $current_project_id, $depth_to_add, $test, $pretty_print_projects, &$rename_counter ) {
		if ( $element_to_add ) {
			$str_log_line = '';

			// If Title is not found, we add it:
			if ( !isset($element_to_add['Title']) || !$element_to_add['Title'] ) {
				$element_to_add['Title'] = 'No name ' . '#REN' . $rename_counter . '#';
				$rename_counter++;
			}

			if ( $pretty_print_projects ) {
				$str_log_line = '*';
				for ( $i=1; $i<=($depth_to_add-1); $i++ ) {
					$str_log_line .= '··';
				}
				$str_log_line .= $element_to_add['Title'];
			} else {
				$str_log_line .= 'password=';
				$str_log_line .= $element_to_add['Title'];
			}

			// Add if not test
			if ( ! $test ) {
				$renamed = '';
				$imported_id = $this->_add_password($element_to_add, $current_project_id, $rename_counter, $renamed);

				if ( $renamed ) {
					$str_log_line .= ' ' . $renamed;					
				}
				$str_log_line .= ' imported_id=' . $imported_id;		
			}

			$this->_write_log($str_log_line);

			// Delete data of the password
			$element_to_add = array();
		} // we have a password to add (element_to_add)
	} // _log_import_password()

	// Adds the password to TPM
	protected function _add_password ( $element_to_add, $project_id, &$rename_counter, &$renamed) {
		$imported_id = 0;
		$arr_rbody = array();
		$s1 = $this->_add_password_basic($element_to_add, $project_id, $imported_id, $arr_rbody);

		if ( $s1 == 201) {
			return $imported_id;
		} else {
			if ( $s1 == 409 && $arr_rbody['message'] == 'This password already exists in the project' ) {
				// Let's add a suffix to the name and try it again
				$renamed = '#REN' . $rename_counter . '#';
				$rename_counter++;
				$element_to_add['Title'] = $element_to_add['Title'] . ' ' . $renamed;

				$imported_id = 0;
				$arr_rbody = array();
				$s1 = $this->_add_password_basic($element_to_add, $project_id, $imported_id, $arr_rbody);

				if ( $s1 == 201) {
					return $imported_id;
				} else {
					$this->_tpm_die('Error adding password: ' . $status . '-' . $arr_rbody['message']);	
				}

			} else {
				$this->_tpm_die('Error adding password: ' . $status . '-' . $arr_rbody['message']);	
			}
		}
	} // _add_password()

	protected function _add_password_basic ( $element_to_add, $project_id, &$imported_id, &$arr_rbody) {
		$apwd = array();
		$apwd['project_id'] = $project_id;
		$imported_id = 0;

		$custom_labels = array();
		$custom_field_idx = 1;
		foreach ( $element_to_add as $eafield => $eavalue ) {
			if ( $eafield != 'UUID' ) {
				$tpm_field = '';
				switch ( $eafield ) {
					case 'Title':
						$tpm_field = 'name';
						break;
					case 'URL':
						$tpm_field = 'access_info';
						break;
					case 'UserName':
						$tpm_field = 'username';
						break;
					case 'Email':
						$tpm_field = 'email';
						break;
					case 'Password':
						$tpm_field = 'password';
						break;
					case 'expiry_date':
						$tpm_field = 'expiry_date';
						break;
					case 'Notes':
						$tpm_field = 'notes';
						break;
					case 'Tags':
						$tpm_field = 'tags';
						break;
					default:						
						if ( $custom_field_idx <= 10 ) { // TPM can only have 10 custom fields
							$tpm_field = 'custom_data' . $custom_field_idx;
							$custom_labels[$custom_field_idx] = $eafield;
							$custom_field_idx++;	
						}
						break;
				}
				if ( $tpm_field ) {
					$apwd[$tpm_field] = $eavalue;		
				}
			} // do not import UUID
		} // foreach

		$rb = json_encode($apwd);
		$ret = $this->_http_post('passwords.json', $rb);

		$status = intval($ret[0]);
		$arr_rbody = json_decode($ret[1], TRUE);

		if ( $status == 201 ) {
			// Added, get the ID
			$imported_id = intval($arr_rbody['id']);

			// Custom fields labels?
			if ( $custom_labels ) {
				$acf = array();
				foreach ($custom_labels as $clidx => $cldat) {
					if ( $clidx > 10 ) { // TPM can only have 10 custom fields
						break;
					}
					$acf['custom_label' . $clidx] = $cldat;
					$acf['custom_type' . $clidx] = 'text';
				}	

				if ( $acf ) {
					$acf_enc = json_encode($acf);
					$retcf = $this->_http_put('passwords/' . $imported_id . '/custom_fields.json', $acf_enc);	
					// Log if status is not 204
					$cfstatus = intval($retcf[0]);					
					if ( $cfstatus != 204 ) {
						$cfarr_rbody = json_decode($cfret[1], TRUE);
						$this->_write_log('Error updating custom fields definitions of the password below: ' . $cfstatus . '-' . $cfarr_rbody['message']);	
					}
				}
			}
		}

		return $status;
	} // _add_password_basic()

	// **************** XML FILE ****************

	protected function _open_xml ( ) {
		if ( $this->_file ) {
			$this->_reader = new XMLReader();
			if ( ! $this->_reader->open($this->_file) ) {
			    $this->_tpm_die('Failed to open XML file: ' . $this->_file);
			}
		} else {
			$this->_tpm_die('No XML file.');
		}
	}

	protected function _close_xml ( ) {
		if ( $this->_reader ) {
			$this->_reader->close();
		}
	}	

	// **************** API ****************

	protected function _http_post ( $request_uri, $request_body_json) {
		// Construct headers
		$headers = array(
			'Content-Type: application/json; charset=utf-8',
			'Expect:' // workaround to the 100-continue header
		);

		// Make request
		$ch = curl_init($this->_tpm_base_url . 'api/v' . self::API_VERSION . '/' . $request_uri);
		curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
		curl_setopt($ch, CURLOPT_HEADER, TRUE); // Includes the header in the output
		curl_setopt($ch, CURLOPT_POST, TRUE);
		curl_setopt($ch, CURLOPT_POSTFIELDS, $request_body_json);
		curl_setopt($ch, CURLOPT_USERPWD, $this->_tpm_username . ":" . $this->_tpm_password); // HTTP Basic Authentication

		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
		
		$result = curl_exec($ch);	
		$status = curl_getinfo($ch, CURLINFO_HTTP_CODE);
		curl_close($ch);
		
		// Separate response headers and body
		list($headers, $body) = explode("\r\n\r\n", $result, 2);
		
		return array($status, $body);
	} // http_post()

	protected function _http_put ( $request_uri, $request_body_json) {
		// Construct headers
		$headers = array(
			'Content-Type: application/json; charset=utf-8',
			'Expect:' // workaround to the 100-continue header
		);

		$ch = curl_init($this->_tpm_base_url . 'api/v' . self::API_VERSION . '/' . $request_uri);
		curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
		curl_setopt($ch, CURLOPT_HEADER, TRUE); // Includes the header in the output
		curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "PUT");
		curl_setopt($ch, CURLOPT_POSTFIELDS, $request_body_json);
		curl_setopt($ch, CURLOPT_USERPWD, $this->_tpm_username . ":" . $this->_tpm_password); // HTTP Basic Authentication

		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
		
		$result = curl_exec($ch);	
		$status = curl_getinfo($ch, CURLINFO_HTTP_CODE);
		curl_close($ch);
		
		// Separate response headers and body
		list($headers, $body) = explode("\r\n\r\n", $result, 2);
		
		return array($status, $body);
	} // _http_put()

	// **************** OTHER ****************

	protected function _tpm_die ( $strm ) {
		$this->_write_log('Dying: ' . $strm);
		$this->_close_log();
		$this->_close_xml();
		die($strm);
	} 
	
} // class TPM_Keepass_Import